# ===================================================
# ===================================================
# Relevant functions for empirical application
# ===================================================
# ===================================================



# ===================================================
# Wrapper function for VaR comparison
# ===================================================


# Inputs: 
# c: specifies the stock (1=AAPL, 2=DIS, 3=MSFT, 4=NKE or 5=PFE)
# n: length of rolling window
# h: choice of scoring function for evaluation (h=1 is standard tick loss, h=0 alternative loss function)


VaR.comparison <- function(c, n, h){    # h=1: 1-homogenous scoring function (standard), h=0: 0-homogenous scoring function
    # 1. Prepare data
   Results <- matrix(NA, nrow=2, ncol=12, 
                     dimnames = list(c("MAX", "NP"), 
                                     c("5% Quantile Score Ratio", "DM", "1% QSR", "DM", "0.5% QSR", "DM", "0.1% QSR", "DM", 
                                       "0.05% QSR", "DM", "0.001% QSR", "DM"))) 
    cat("c = ", c, "\n")
    lr.DAX    <- V[c, ]                 # Investigate log-returns
    N         <- length(lr.DAX)
    
    
    # 2. Forecast CVaR and CES
    
    alpha    <- c(0.05, 0.01, 0.005, 0.001, 0.0005, 0.00001)
    no.alpha <- length(alpha)
    
    CVaR.forecasts.QCRPS <- matrix(NA, nrow = no.alpha, ncol = (N-n))
    CVaR.forecasts.MAX   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
    CVaR.forecasts.NP    <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
    
    k.QCRPS <- rep.int(NA, times=(N-n))
    k.MAX   <- rep.int(NA, times=(N-n)) 
    
    GARCH.param <- matrix(NA, nrow=4, ncol= (N-n), dimnames = list(c("omega.hat", "alpha.hat", "beta.hat", "tail index.hat"), 1:(N-n))) 
    
    
    for(j in n : (N-1) ){
      forecast            <- VaR.forecast(lr.DAX[(j-n+1) : j], alpha, nu = 10)
      
      k.QCRPS[(j-n+1)] <- forecast$k.QCRPS
      k.MAX[(j-n+1)]   <- forecast$k.MAX
      
      GARCH.param[1:3, (j-n+1)] <- forecast$theta.hat        # save GARCH(1,1) parameter estimates
      GARCH.param[4, (j-n+1)]   <- forecast$alpha.hat        # save tail indices implied by GARCH(1,1) parameter estimates
      
      for(a in 1 : no.alpha){
        CVaR.forecasts.QCRPS[a, (j-n+1)] <- forecast$VaR.QCRPS[a]
        CVaR.forecasts.MAX[a, (j-n+1)]   <- forecast$VaR.MAX[a]
        CVaR.forecasts.NP[a, (j-n+1)]    <- forecast$VaR.NP[a] 
      }
    }
    k <- matrix(c(mean(k.MAX), mean(k.QCRPS)), nrow=1, ncol=2, dimnames=list(c(""), c("k.Q", "k.QCRPS")) )
    
    # 3. Comparative Backtest
    
    X <- lr.DAX[(n+1):N]
    
    for(a in 1:no.alpha){
       
        # 5. Diebold Mariano (1995) Test of equal predictive ability
        
        l1.VaR <- loss.VaR(data = X, VaR = CVaR.forecasts.QCRPS[a, ], 1-alpha[a], h)
        l2.VaR <- loss.VaR(data = X, VaR = CVaR.forecasts.MAX[a, ],   1-alpha[a], h)
        l3.VaR <- loss.VaR(data = X, VaR = CVaR.forecasts.NP[a, ],    1-alpha[a], h)
        
        Results[1, 1 + (a-1)*2] <- sum(l1.VaR) / sum(l2.VaR)
        Results[2, 1 + (a-1)*2] <- sum(l1.VaR) / sum(l3.VaR)
        
        # print("VaR:")
        dm.l1l2 <- dm95.test(l1.VaR, l2.VaR, "two.sided")
        dm.l1l3 <- dm95.test(l1.VaR, l3.VaR, "two.sided")
        
        Results[1, 2 + (a-1)*2] <- dm.l1l2
        Results[2, 2 + (a-1)*2] <- dm.l1l3
    }
    
    # 6. Summarize Results in Matrix
    return( list(Results = Results, GARCH.param = GARCH.param, CVaR.forecasts = CVaR.forecasts.MAX[2, ], k=k) )
}



# ===================================================
# Function to generate conditional VaR forecasts
# ===================================================


# Inputs: 
# X: log-losses
# alpha: relevant quantiles
# nu: number of residuals to discard due to initialization effects in variance equation


VaR.forecast <- function(X, alpha, nu){  # VaR forecasts using Chan et al. (2007) and Hoga (2017+)
  n <- length(X) - nu
  no.alpha <- length(alpha)
  
  #####################################
  
  # 1. Fit model to generate conditional risk forecasts
  # 1.1. Estimate GARCH(1,1) model using standard QMLE for log-returns
  outp      <- garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                        include.mean = FALSE, algorithm = "lbfgsb", trace = FALSE)
  param     <- outp@fit$matcoef
  theta.hat <- as.numeric(param[ ,1])


  # 1.2. Compute residuals U.hat
  X         <- X[(nu+1) : (nu+n)]
  sigma.hat <- outp@sigma.t[(nu+1) : (nu+n)]  # discard first nu volatility estimates, due to initialization effects
  U.hat     <- X / sigma.hat
  sort.U    <- sort.int(U.hat, decreasing = TRUE, method = "quick")

  # 1.3 Volatility estimate
  sigma.hat.n1 <- sqrt( theta.hat[1] + theta.hat[2] * X[n]^2 + theta.hat[3] * sigma.hat[n]^2 )
  
  
  # 2. Compute optimal k* using [Dea16+, p. 16] (Tail Index Estimation: Quantile Driven Threshold Selection)
  
  k.max <- floor(n^0.6)
  
  sup.MAX <- rep.int(NA, k.max)
  QCRPS   <- H <- rep.int(0, k.max)
  for(k in 1 : k.max){
    H[k]      <- 1/k * sum(log( sort.U[1 : (k+1)] / sort.U[(k+1)] ))  # Hill estimate
    Q.k       <- (k/(1 : k.max))^( H[k] ) * sort.U[(k+1)]             # Weissman estimate
    deviation <- sort.U[2 : (k.max+1)] - Q.k                          # deviation between empirical and Weissman quantile estimates
    
    sup.MAX[k] <- max( abs(deviation) )                               # Choice of [Dea16+]

    for(i in 1 : k.max){
      QCRPS[k] <- QCRPS[k] + sum( (1*(sort.U <= Q.k[i]) - (1 - i/n)) * (Q.k[i] - sort.U) )
    }
  }
   # 3. Compute Risk Measure Forecasts
  k.MAX   <- which.min( sup.MAX )
  k.QCRPS <- which.min( QCRPS )
   
  H.MAX   <- sort.U[k.MAX+1]   * ((n*alpha) / k.MAX)^(-H[k.MAX])     # H EQ estimator
  H.QCRPS <- sort.U[k.QCRPS+1] * ((n*alpha) / k.QCRPS)^(-H[k.QCRPS]) # H EQ estimator
  H.NP    <- rep.int(NA, no.alpha)                  
  for(a in 1:no.alpha){
    H.NP[a]       <- sort.U[floor(n*alpha[a])+1]
  }
  
  CVaR.MAX   <- sigma.hat.n1 * H.MAX     
  CVaR.QCRPS <- sigma.hat.n1 * H.QCRPS   
  CVaR.NP    <- sigma.hat.n1 * H.NP      
  
  
  return( list( VaR.MAX = c(CVaR.MAX), VaR.QCRPS = c(CVaR.QCRPS), VaR.NP = c(CVaR.NP),
                k.MAX = c(k.MAX), k.QCRPS = c(k.QCRPS),
                theta.hat = c(theta.hat), alpha.hat = 1 / H[k.MAX]) )
}




# ===================================================
# Scores for VaR forecasts
# ===================================================


# Inputs: 
# data: verifying observations
# VaR: VaR forecasts
# alpha: relevant quantile levels
# h: h=1 corresponds to tick loss (or also: quantile score), h=0 to alternative score


loss.VaR <- function(data, VaR, alpha, h){  
   S <- numeric(length(VaR))
   if(h==1){
      S <- (data - VaR) * (alpha - 1*(data <= VaR)) 
   }
   # ind1 = (data>VaR) & (VaR>0)
   # ind2 = (data<VaR) & (VaR>0)
   if(h==0){
      S <- ( exp(data) / (1+exp(data)) - exp(VaR) / (1+exp(VaR)) ) * (alpha - 1*(data <= VaR))             
   }
  return(S)
}




# ===================================================
# Diebold-Mariano (1995) test of equal predicitve ability
# ===================================================


# Inputs: 
# loss1: losses using VaR forecasts 1
# loss2: losses using VaR forecasts 2
# alternative: type of test
# h: number of lags to include in long-run variance estimation



dm95.test <- function (loss1, loss2, alternative = c("two.sided", "less", "greater"), h = 1){
  alternative <- match.arg(alternative)
  d <- c(loss1) - c(loss2)
  d.cov <- acf(d, na.action = na.omit, lag.max = h - 1, type = "covariance",
               plot = FALSE)$acf[, , 1]
  d.var <- sum(c(d.cov[1], 2 * d.cov[-1]))/length(d)
  dv <- d.var
  if (dv > 0)
    STATISTIC <- mean(d, na.rm = TRUE)/sqrt(dv)
  else if (h == 1)
    stop("Variance of DM statistic is zero")
  else {
    warning("Variance is negative, using horizon h=1")
    return(dm95.test(loss1, loss2, alternative, h = 1))
  }
  n <- length(d)
  k <- ((n + 1 - 2 * h + (h/n) * (h - 1))/n)^(1/2)
  STATISTIC <- STATISTIC * k
  names(STATISTIC) <- "DM"
  if (alternative == "two.sided")
    PVAL <- 2 * pt(-abs(STATISTIC), df = n - 1)
  else if (alternative == "less")
    PVAL <- pt(STATISTIC, df = n - 1)
  else if (alternative == "greater")
    PVAL <- pt(STATISTIC, df = n - 1, lower.tail = FALSE)
  return(PVAL)
}


